#include "lighting.h"
#include "simple_list.h"
#include <assert.h>
#include <stdlib.h>
#include <vector>
#include <algorithm>
#include "vec4.h"
#include "vec3.h"
#include "mat3d.h"

void validate_shine_table( struct Rcontext *ctx, uint32 side, float shininess )
{
	struct shine_tab *list = ctx->_ShineTabList;
	struct shine_tab *s;

	assert(side < 2);

	foreach(s, list)
		if ( s->shininess == shininess )
			break;

	if (s == list) {
		int j;
		float *m;

		foreach(s, list)
			if (s->refcount == 0)
				break;

		m = s->tab;
		m[0] = 0.0;
		if (shininess == 0.0) {
			for (j = 1 ; j <= SHINE_TABLE_SIZE ; j++)
				m[j] = 1.0;
		}
		else {
			for (j = 1 ; j < SHINE_TABLE_SIZE ; j++) {
				double t, x = j / (float) (SHINE_TABLE_SIZE - 1);
				if (x < 0.005) /* underflow check */
					x = 0.005;
				t = pow(x, shininess);
				if (t > 1e-20)
					m[j] = (float) t;
				else
					m[j] = 0.0;
			}
			m[SHINE_TABLE_SIZE] = 1.0;
		}

		s->shininess = shininess;
	}

	if (ctx->_ShineTable[side])
		ctx->_ShineTable[side]->refcount--;

	ctx->_ShineTable[side] = s;
	move_to_tail( list, s );
	s->refcount++;
}

/*
 * Whenever the spotlight exponent for a light changes we must call
 * this function to recompute the exponent lookup table.
 */
/* FIXME These should be static. */
static void invalidate_spot_exp_table( Light *l )
{
	l->_SpotExpTable[0][0] = -1;
}

#define FLT_MIN (1.0e-37)
static void validate_spot_exp_table( Light *l )
{
	int i;
	double exponent = l->SpotExponent;
	double tmp = 0;
	int clamp = 0;

	l->_SpotExpTable[0][0] = 0.0;

	for (i = EXP_TABLE_SIZE - 1; i > 0 ;i--) {
		if (clamp == 0) {
			tmp = pow(i / (double) (EXP_TABLE_SIZE - 1), exponent);
			if (tmp < FLT_MIN * 100.0) {
				tmp = 0.0;
				clamp = 1;
			}
		}
		l->_SpotExpTable[i][0] = (float) tmp;
	}
	for (i = 0; i < EXP_TABLE_SIZE - 1; i++) {
		l->_SpotExpTable[i][1] = (l->_SpotExpTable[i+1][0] -
				l->_SpotExpTable[i][0]);
	}
	l->_SpotExpTable[EXP_TABLE_SIZE-1][1] = 0.0;
}


#define MALLOC_STRUCT(T) (struct T*)malloc(sizeof(struct T))

static void init_light(Light *l) {
	l->SpotExponent = 0.0;
	l->flags = 0;
	l->ambient = Color3(0,0,0);
	l->quad_attenuation(0);
	// set default diffuse, ambient, etc.
	invalidate_spot_exp_table(l);
}

void light_spot_exponent(Light *l, float exp) {
	l->SpotExponent = exp;
	validate_spot_exp_table(l);
}

void light_spot_cutoff(Light *l, float angle) {
	if (angle == 180.0f) {
		// 180 degree spotlight is not a spotlight
		l->flags &= ~LIGHT_SPOT;
	} else {
		assert_msg(angle >= 0.0 && angle <= 90.0f, "invalid spotlight angle");
		l->spot_cutoff = fcos(angle * DEG2RAD);
		l->flags |= LIGHT_SPOT;
	}
}

// spot direction (light normal) in eye coords (not automatically normalized)
void light_spot_direction_eye(Light *l, Vector4 *dir) {
	l->norm.x = dir->x;
	l->norm.y = dir->y;
	l->norm.z = dir->z;
	l->norm.w = 0;
	l->norm.normalize_self();
}

void light_pos_eye(Light& l, const Vec4& v) {
	l.pos = v;
	if (l.pos.w != 0.0f) 
		l.flags |= LIGHT_POSITIONAL;
}
void light_pos(Light& l, const Vec4& v) {
	Vec4 E = v;
	plx_mat_identity();  // really should be a plx_mat3d_load() function
	plx_mat3d_apply(PLX_MAT_MODELVIEW);
	E.transform_nodiv();
	light_pos_eye(l, E); // slightly inefficient
}

void light_diffuse(Light *l, const Color3 &c) {
	l->diffuse = c;
}

void light_specular(Light *l, const Color3 &c) {
	l->specular = c;
}

void light_ambient(Light *l, const Color3 &c) {
	l->ambient = c;
}

void light_enable_specular(Light *l) {
	l->flags |= LIGHT_SPECULAR;
}

void light_enable(struct Rcontext *ctx, Light *l) {
	assert_msg(!(l->flags & LIGHT_ENABLED), "light already enabled");
	l->flags |= LIGHT_ENABLED;
	ctx->lighting.enabled_lights.push_back(l);
}

void light_disable(struct Rcontext *ctx, Light *l) {
	assert_msg(l->flags & LIGHT_ENABLED, "light not enabled");
	l->flags &= ~LIGHT_ENABLED;
	typedef std::vector<Light *> lightvec;
	lightvec& list = ctx->lighting.enabled_lights;
	lightvec::iterator pos = std::find(list.begin(), list.end(), l);
	list.erase(pos);
}

Light *light_new(struct Rcontext *ctx) {
	static int num_enabled = 0;
	assert_msg(num_enabled < MAX_LIGHTS, "too many lights");
	return &(ctx->lighting.light[num_enabled++]);
}

void init_lighting(struct Rcontext *ctx) {
	int i;

	for (i = 0; i < MAX_LIGHTS; i++)
		init_light(&ctx->lighting.light[i]);
	ctx->lighting.enabled_lights.reserve(MAX_LIGHTS);

	/* Imported from Mesa 6.0 */
	ctx->_ShineTabList = MALLOC_STRUCT( shine_tab );
	make_empty_list( ctx->_ShineTabList );
	/* Allocate 10 (arbitrary) shininess lookup tables */
	for (i = 0 ; i < 10 ; i++) {
		struct shine_tab *s = MALLOC_STRUCT( shine_tab );
		s->shininess = -1;
		s->refcount = 0;
		insert_at_tail( ctx->_ShineTabList, s );
	}
	// FIXME set default global ambient (and flag?)
	ctx->lighting.base_color = Color3(0, 0, 0);
	ctx->lighting.local_viewer(false);
	// ctx->lighting.global_ambient_ = Color3(0.2, 0.2, 0.2);
}

void shutdown_lighting(struct Rcontext *ctx) {
	// free
}

/* Temporary matrices. */
#define align32 __attribute__((aligned(32)))
static matrix_t align32 mat_out;
static matrix_t align32 mat_in;

/* Call this function whenever the modelview matrix changes; it transforms
 * the internal lighting vectors, stored in eye coordinates, into the
 * corresponding _obj vectors in object coordinates. */
void Lighting::update() {
	// Invert modelview matrix and load into xmtrx
	plx_mat_identity();
	mat_store(&mat_out);
	plx_mat3d_apply(PLX_MAT_MODELVIEW);
	plx_mat3d_store(&mat_in);
	uint32 flags = MAT_FLAG_ROTATION | MAT_FLAG_TRANSLATION;
	invert_matrix_3d(&mat_out, &mat_in, flags);
	mat_load(&mat_out);

	// Transform infinite viewer direction into object coords (for specular)
	Vec4 inf_viewer(0, 0, 1, 0);
	inf_viewer.transform_nodiv();
	this->inf_viewer_obj = inf_viewer;
	// Transform local viewer position into object coords (for specular)
	//if (local_viewer()) {
		Vec4 local_viewer(0, 0, 0, 1);
		local_viewer.transform_nodiv();
		this->local_viewer_obj = local_viewer;
	//}
	
	// Transform light vectors from eye to object coordinates.
	typedef std::vector<Light *>::const_iterator light_iter;
	
	for (light_iter pos = enabled_lights.begin(); pos != enabled_lights.end(); ++pos) {
		Light *l = *pos;
		// Transform light position or direction into object coordinates
		// light.w is important!  The translation is scaled by the w value
		// (which should be 0 or 1)
		l->pos_obj.x = l->pos.x; l->pos_obj.y = l->pos.y; l->pos_obj.z = l->pos.z;
		mat_trans_nodiv(l->pos_obj.x, l->pos_obj.y, l->pos_obj.z, l->pos.w);
		// Transform spot direction into object coords (for spotlight)
		l->norm_obj.x = l->norm.x; l->norm_obj.y = l->norm.y; l->norm_obj.z = l->norm.z;
		mat_trans_nodiv(l->norm_obj.x, l->norm_obj.y, l->norm_obj.z, l->norm.w);
		// Note: spot_dir doesn't need to be normalized unless the modelview matrix does
		// a non-uniform scaling, as it's a normal (direction)
		// normalize(&spot_dir);
		
		// For directional lights, direction to each vertex is constant
		if (!(l->flags & LIGHT_POSITIONAL)) {
			l->pos_obj.normalize_self();
		}
	}
}

static inline void light_enable_global_ambient() {
	// FIXME Cannot enable/disable global ambient right now.
}


void Lighting::precompute_material_products() {
	typedef std::vector<Light *>::const_iterator light_iter;
	for (light_iter pos = enabled_lights.begin(); pos != enabled_lights.end(); ++pos) {
		Light *l = *pos;
		l->mat_diffuse = material_.diffuse() * l->diffuse;
		l->mat_specular = material_.specular() * l->specular;
		l->mat_ambient = material_.ambient() * l->ambient;
	}
	// Miraculously, enabling this line causes a major performance hit,
	// even though it's only called once.  Actually -setting- global
	// ambient makes it even worse!
	// base_color = material_.emissive() + global_ambient_;
	// Either of these are okay, but as soon as you set global_ambient_
	// from outside, it slows down.  I suspect a cache issue or
	// major gcc optimization problem.  Note that a change here
	// (changing this object file only) can affect the loop in
	// light.o -- so something is seriously wrong.
	// base_color = global_ambient_;
	base_color = material_.emissive();
}
