#ifndef __LIGHTING_H
#define __LIGHTING_H

#include <dc/pvr.h>
#include <vector>
#include <math.h>
#include "math.h"
#include "color3.h"
#include "vec4.h"

/* Imported from Mesa 6.0 */
/**
 * Material shininess lookup table.
 */
/* Mesa uses 2 active shininess tables [front and back] which are just
 * pointers into an MRU shininess table list. */
#define EXP_TABLE_SIZE 512
#define SHINE_TABLE_SIZE 256
#define MAX_LIGHTS 8

typedef struct {
	float a, r, g, b;
} Color4;

struct shine_tab {
   struct shine_tab *next, *prev;
   float tab[SHINE_TABLE_SIZE+1];
   float shininess;
   uint32 refcount;
};

#define LIGHT_POSITIONAL 0x1
#define LIGHT_SPOT 0x2
#define LIGHT_SPECULAR 0x4
#define LIGHT_ENABLED 0x8
#define LIGHT_ATTENUATED 0x10

#define LIGHT_MODEL_GLOBAL_AMBIENT 0x1
#define LIGHT_MODEL_LOCAL_VIEWER 0x2

struct Light {
	float _SpotExpTable[EXP_TABLE_SIZE][2];
	float SpotExponent;
	float spot_cutoff;
	float quad_att; // Quadratic attenuation
	Color3 ambient, diffuse, specular;
	Color3 mat_ambient, mat_diffuse, mat_specular; // Precomputed light*material values.
	/* The light object vectors can be Vec3s, because w is never used. */
	Vec4 pos, norm;
	Vec3 pos_obj, norm_obj;
	uint32 flags;
public:
	void quad_attenuation(float att) { 
		quad_att = att; 
		if (att == 0) flags &= ~LIGHT_ATTENUATED; 
		else flags |= LIGHT_ATTENUATED;
	}
};

class Material {
	Color3 _ambient, _specular, _diffuse, _emissive;
	float _alpha;
public:
//	Constructor problems; the _emissive line causes major
//	performance hit; the _diffuse line has no effect -unless-
//	mat.emissive(Color3) is called in main().
//	Material() : _emissive(Color3(0,0,0)) {}
//  Material() : _diffuse(Color3(0,0,0)) {}
	void ambient(const Color3& c) { _ambient = c; }
	void specular(const Color3& c) { _specular = c; }
	void diffuse(const Color3& c) { _diffuse = c; }
	void emissive(const Color3& c) { _emissive = c; }
	void alpha(float a) { _alpha = a; }
	const Color3& ambient() const { return _ambient; }
	const Color3& emissive() const { return _emissive; }
	const Color3& diffuse() const { return _diffuse; }
	const Color3& specular() const { return _specular; }
	float alpha() const { return _alpha; }
};

class Lighting {
public:
	Light light[MAX_LIGHTS];
	Color3 base_color;
	// Color3 global_ambient_;
	std::vector<Light *> enabled_lights;
	Vec3 inf_viewer_obj;
	Vec3 local_viewer_obj;
	Material material_;
	uint32 flags;
public:
	void update();
	void precompute_material_products();
	void set_material(Material &m) {
		material_ = m;
		precompute_material_products();
	}
	void local_viewer(bool v) {
		if (v) flags |= LIGHT_MODEL_LOCAL_VIEWER;
		else   flags &= ~LIGHT_MODEL_LOCAL_VIEWER;
	}
	bool local_viewer() { return flags & LIGHT_MODEL_LOCAL_VIEWER; }
	const Material& material() {
		return material_;
	}
	/*  Having major problems with this performance-wise.
	void global_ambient(const Color3& c) {
		printf("setting global ambient...\n");
		global_ambient_ = c;
	}
	*/

//	int num_enabled;
};

struct Rcontext {
	struct shine_tab *_ShineTabList;
	struct shine_tab *_ShineTable[2];
	Lighting lighting;
	pvr_dr_state_t	dr;
};

/* Imported from Mesa 6.0 */
/* Lerp between adjacent values in the f(x) lookup table, giving a
 * continuous function, with adequeate overall accuracy.  (Though
 * still pretty good compared to a straight lookup).
 * Result should be a float.
 */
#define GET_SHINE_TAB_ENTRY( table, dp, result )			\
do {									\
   struct shine_tab *_tab = table;					\
   float f = (dp * (SHINE_TABLE_SIZE-1));				\
   int k = (int) f;							\
   if (k > SHINE_TABLE_SIZE-2) 						\
      result = (float) pow( dp, _tab->shininess );		\
   else									\
      result = _tab->tab[k] + (f-k)*(_tab->tab[k+1]-_tab->tab[k]);	\
} while (0)

#define GET_SPOT_EXP_ENTRY( L, dp, result ) \
do { \
	  double x = dp * (EXP_TABLE_SIZE-1); \
	  int k = (int)x; \
	  result = (float) ((L)->_SpotExpTable[k][0] + (x-k)*(L)->_SpotExpTable[k][1]); \
} while (0)

void init_lighting(struct Rcontext *ctx);
void shutdown_lighting(struct Rcontext *ctx);
void validate_shine_table( struct Rcontext *ctx, uint32 side, float shininess );
void light_spot_exponent(Light *l, float exp);
void light_diffuse(Light *l, const Color3& c);
void light_specular(Light *l, const Color3& c);
void light_ambient(Light *l, const Color3& c);
void light_enable_specular(Light *l);
void light_pos_eye(Light& l, const Vec4& v);
void light_pos(Light& l, const Vec4& v);
void light_spot_direction_eye(Light *l, Vector4 *dir);
void light_spot_cutoff(Light *l, float angle);
void light_enable(struct Rcontext *ctx, Light *l);
void light_disable(struct Rcontext *ctx, Light *l);
//void light_enable_global_ambient();
Light *light_new(struct Rcontext *ctx);

#endif /* __LIGHTING_H */
