#ifndef __LIGHT_MATH_H
#define __LIGHT_MATH_H

#include <plx/matrix.h>
#ifdef _arch_dreamcast
# include <dc/fmath.h>
#else
# include <math.h>
#endif

typedef struct {
	float x, y, z, w;
} Vector4;

#define DEG2RAD (F_PI / 180.0f)

/* Floating point inner product (dot product) */
#define __fipr(x, y, z, w, a, b, c, d) ({ \
	register float __x __asm__("fr0") = (x); \
	register float __y __asm__("fr1") = (y); \
	register float __z __asm__("fr2") = (z); \
	register float __w __asm__("fr3") = (w); \
	register float __a __asm__("fr4") = (a); \
	register float __b __asm__("fr5") = (b); \
	register float __c __asm__("fr6") = (c); \
	register float __d __asm__("fr7") = (d); \
	__asm__ __volatile__( \
		"fipr   fv4,fv0" \
		: "+f" (__w) \
		: "f" (__x), "f" (__y), "f" (__z), "f" (__w), "f" (__a), "f" (__b), "f" (__c), "f" (__d) \
		); \
		__w; })

extern inline float fipr(float x, float y, float z, float w,
		float a, float b, float c, float d) {
	return __fipr(x, y, z, w, a, b, c, d);
}

/* Square of vector magnitude */
#define __fipr_magnitude_sqr(x, y, z, w) ({ \
	register float __x __asm__("fr4") = (x); \
	register float __y __asm__("fr5") = (y); \
	register float __z __asm__("fr6") = (z); \
	register float __w __asm__("fr7") = (w); \
	__asm__ __volatile__( \
		"fipr   fv4,fv4" \
		: "+f" (__w) \
		: "f" (__x), "f" (__y), "f" (__z), "f" (__w) \
		); \
		__w; })

extern inline float magnitude_sqr(float x, float y, float z) {
	return __fipr_magnitude_sqr(x, y, z, 0.0f);
}

/* Transform vector, without any perspective division. */
#define mat_trans_nodiv(x, y, z, w) { \
	register float __x __asm__("fr0") = (x); \
	register float __y __asm__("fr1") = (y); \
	register float __z __asm__("fr2") = (z); \
	register float __w __asm__("fr3") = (w); \
	__asm__ __volatile__( \
		"ftrv	xmtrx,fv0\n" \
		: "=f" (__x), "=f" (__y), "=f" (__z), "=f" (__w) \
		: "0" (__x), "1" (__y), "2" (__z), "3" (__w) ); \
	x = __x; y = __y; z = __z; w = __w; \
}

/* Inverse square root.  Mesa's version is a
 * little slower than the SH-4 fsqrt and fdiv. */
#define inv_sqrtf(x) (frsqrt(x))
// #define inv_sqrtf(x) (1.0f / fsqrt(x))
// #define inv_sqrtf(x) _mesa_inv_sqrtf(x)

/* Imported from Mesa 6.0 */
static inline float _mesa_inv_sqrtf(float n)
{
        float r0, x0, y0;
        float r1, x1, y1;
        float r2, x2, y2;
        union { float f; unsigned int i; } u;
        unsigned int magic;

        magic = 381 << 23;
        magic -= (int)(0.0332281 * (1 << 25));

        u.f = n;
        u.i = (magic - u.i) >> 1;

        x0 = 1.0f;
        y0 = 0.5f * n;
        r0 = u.f;

        x1 = x0 * r0;
        y1 = y0 * r0 * r0;
        r1 = 1.5f - y1;
		// return x1 * r1;      // trade a little speed for (lots of?) precision here

        x2 = x1 * r1;
        y2 = y1 * r1 * r1;
        r2 = 1.5f - y2;

        return x2 * r2;
}

#ifdef _arch_dreamcast
inline float SQRT(float x) { return fsqrt(x); }
inline float RSQRT(float x) { return frsqrt(x); }
#else
inline float SQRT(float x) { return (float)sqrt(x); }
inline float RSQRT(float x) { return 1.0f / frsqrt(x); }
#endif

static inline void normalize(Vector4 *v) {
	// float len = inv_sqrtf(magnitude_sqr(v->x, v->y, v->z));
	// It's slower to use fipr here [or in dot()]
	float len = inv_sqrtf(v->x * v->x + v->y * v->y + v->z * v->z);
	v->x *= len; v->y *= len; v->z *= len;
}
static inline float dot(Vector4 *u, Vector4 *v) {
    return (u->x * v->x)
        + (u->y * v->y)
        + (u->z * v->z);
}

static inline void normalize_asm(Vector4 *v) {
	float len = inv_sqrtf(magnitude_sqr(v->x, v->y, v->z));
	//float len = inv_sqrtf(magnitude_sqr(v->x, v->y, v->z));
	//if (mag < 0) { printf("mag bug\n"); }
	//if (len < 0) { printf("len bug\n"); }
	v->x *= len; v->y *= len; v->z *= len;
}

static inline float dot_asm(Vector4 *u, Vector4 *v) {
	return fipr(u->x, u->y, u->z, 0.0f, v->x, v->y, v->z, 0.0f);
}

#endif /* __LIGHT_MATH_H */
