#include <kos.h>
#include <vector>
#include <plx/matrix.h>
#include <plx/prim.h>
#include "lighting.h"
#include "model.h"
#include "global.h"

void draw_model(const Model& m) {
	VectorF *v, *v_obj;
	Vec3 *n;
	size_t size = m.nvert;
	Vec3 L; // Light vector
	float d;
	Color3 base = rctx.lighting.base_color;  // Currently a tad faster to make this a 
	                                         // local variable as opposed to an alias.
	bool local_viewer = rctx.lighting.local_viewer();
	Vec3& local_viewer_obj = rctx.lighting.local_viewer_obj;
	Vec3 VD;  // Viewpoint direction (direction from vertex to viewpoint)
	if (!local_viewer)
		VD = rctx.lighting.inf_viewer_obj;

	Light *l;

	n = m.normals;     assert(n);
	v_obj = m.data;	    assert(v_obj);
	v = m.data_xform;	    assert(v);

	pvr_prim(&hdr, sizeof(hdr));
	pvr_dr_init(rctx.dr);

	// Initialize xmtrx with the screenspace transform
	plx_mat_identity();
	plx_mat3d_apply(PLX_MAT_SCREENVIEW);
	plx_mat3d_apply(PLX_MAT_PROJECTION);
	plx_mat3d_apply(PLX_MAT_MODELVIEW);
	// Transform model from object space to screen space
	mat_transform(v_obj, v, size, 16);

	uint32 a0 = plx_pack_alpha(rctx.lighting.material().alpha());   

	typedef std::vector<Light *> enabled_type;
	typedef enabled_type::const_iterator light_iter;
	enabled_type& lights = rctx.lighting.enabled_lights;

	// Disallow optimization based on light types for now
#if 0
	if (!(l->flags & (LIGHT_POSITIONAL|LIGHT_SPECULAR|LIGHT_SPOT))) {
		// Optimized case for 1 directional light only
		draw_light_directional_only(&dr, size, v, n, &L, &diffuse);
		frame++;
		return;
	}
#endif

	while(size--) {
		uint32 vindex = v_obj->flags;
		uint32 lit_color = m.lit_color[vindex];
		if (lit_color != 0) {
			// Cached lighting color was valid, so draw vertex and continue.
			plx_vert_ind(&rctx.dr, v->flags, v->x, v->y, v->z, lit_color);
			n++; v_obj++; v++;
			continue;
		}
		
		// Otherwise, compute lighting for this vertex.

		Color3 sum = base; // set emission and global ambient
		
		for (light_iter pos = lights.begin(); pos != lights.end(); ++pos) {
			float r, g, b;
			float attenuation = 1.0f;
			l = *pos;

			if (l->flags & LIGHT_POSITIONAL) {
				// Positional lights compute direction to each vertex.
				// The normalize slows us down a lot.
				L.x = l->pos_obj.x - v_obj->x;
				L.y = l->pos_obj.y - v_obj->y;
				L.z = l->pos_obj.z - v_obj->z;
				L.normalize_self();
				if (l->flags & LIGHT_SPOT) {
					float d = -L.dot(l->norm_obj);
					if (d <= l->spot_cutoff)
						continue;                 // No contribution from this light.
					GET_SPOT_EXP_ENTRY(l, d, attenuation);
					// This helps for cutoffs that lie outside the
					// maximum attenuation. Experimentally 0.01 is reasonable.
					if (attenuation < 0.01)
						continue;
				}
			} else {
				// Directional light; direction is constant.
				L = l->pos_obj;
			}
			d = L.dot(*n);

			if (d >= 0) {
				// set diffuse light vertex values -- 
				// not currently faster to skip multiplies when d<0
				r = l->mat_diffuse.r * d;
				g = l->mat_diffuse.g * d;
				b = l->mat_diffuse.b * d;
			} else {
				r = 0; g = 0; b = 0; 
			}

			if (l->flags & LIGHT_SPECULAR) {
				// Normalize sum of unit light and unit viewer vector.
				// FIXME normalize need only be done once if light is directional
				if (0 && local_viewer) {  // XXX works (I think) but disabled for now for speed
					// Local viewer: get unit viewer vector
					VD.x = local_viewer_obj.x - v_obj->x;
					VD.y = local_viewer_obj.y - v_obj->y;
					VD.z = local_viewer_obj.z - v_obj->z;
					VD.normalize_self();
				}
				Vec3 s = L + VD; // specular sum
				s.normalize_self();
				d = s.dot(*n);
				if (d > 0.0f) {
					GET_SHINE_TAB_ENTRY(rctx._ShineTable[0], d, d);
					r += l->mat_specular.r * d;
					g += l->mat_specular.g * d;
					b += l->mat_specular.b * d;
				}
			}
			r += l->mat_ambient.r;
			g += l->mat_ambient.g;
			b += l->mat_ambient.b;
			if (attenuation != 1.0f)
				r *= attenuation; g*= attenuation; b*=attenuation;
			sum.r += r; sum.g += g; sum.b += b;
		}

		sum.clamp();                           // Clamp final rgb values to 1.0
		lit_color = a0 | plx_pack_color3(sum);
		m.lit_color[vindex] = lit_color;    // Cache this vertex's color.

		// Draw the vertex.
		plx_vert_ind(&rctx.dr, v->flags, v->x, v->y, v->z, lit_color);
		n++; v_obj++; v++;
	}

	m.invalidate_lighting();
	//assert(valid_colors == sph->ndups);

	//frame++;
}


void draw_model(const QuadModel& m) {
	Vec3 *v, *v_obj;
	Vec3 *n;
	//size_t size = m.nvert;
	size_t faces = m.nfaces;
	Vec3 L; // Light vector
	float d;
	QuadModel::facedata *f; // Vertices for each face
	Color3 base = rctx.lighting.base_color;  // Currently a tad faster to make this a 
	                                         // local variable as opposed to an alias.
	bool local_viewer = rctx.lighting.local_viewer();
	Vec3& local_viewer_obj = rctx.lighting.local_viewer_obj;
	Vec3 VD;  // Viewpoint direction (direction from vertex to viewpoint)
	if (!local_viewer)
		VD = rctx.lighting.inf_viewer_obj;

	Light *l;

	n = m.normals;     assert(n);
	v_obj = m.data;	    assert(v_obj);
	v = m.data_xform;	    assert(v);
	f = m.faces; assert(f);

	pvr_prim(&hdr, sizeof(hdr));
	pvr_dr_init(rctx.dr);

	// Initialize xmtrx with the screenspace transform
	plx_mat_identity();
	plx_mat3d_apply(PLX_MAT_SCREENVIEW);
	plx_mat3d_apply(PLX_MAT_PROJECTION);
	plx_mat3d_apply(PLX_MAT_MODELVIEW);
	// Transform model from object space to screen space
	//mat_transform(v_obj, v, size, 12);

	uint32 a0 = plx_pack_alpha(rctx.lighting.material().alpha());   

	typedef std::vector<Light *> enabled_type;
	typedef enabled_type::const_iterator light_iter;
	enabled_type& lights = rctx.lighting.enabled_lights;

	while(faces--) {   // foreach face
		// discard backfacing polys
		if (f->pn - local_viewer_obj.dot(*n) > 0) {
			n++; f++;
			continue;
		}
		for (int vtx = 0; vtx < m.nfacevtx; ++vtx) {
			uint32 idx = f->vtx[vtx];
			Vec3 *v_obj = &m.data[idx];
			//Vec3 *v     = &m.data_xform[idx];

			Color3 sum = base; // set emission and global ambient
			
			for (light_iter pos = lights.begin(); pos != lights.end(); ++pos) {
				float r, g, b;
				float attenuation = 1.0f;
				l = *pos;

				if (l->flags & LIGHT_POSITIONAL) {
					// Positional lights compute direction to each vertex.
					// The normalize slows us down a lot.
					L.x = l->pos_obj.x - v_obj->x;
					L.y = l->pos_obj.y - v_obj->y;
					L.z = l->pos_obj.z - v_obj->z;
					float len2 = L.length_2();
					//L.normalize_self();
					L *= RSQRT(len2);
					if (l->flags & LIGHT_SPOT) {
						float d = -L.dot(l->norm_obj);
						if (d <= l->spot_cutoff)
							continue;                 // No contribution from this light.
						GET_SPOT_EXP_ENTRY(l, d, attenuation);
						if (attenuation < 0.01)
							continue;
					}
					attenuation /= 1 + l->quad_att*len2;
					// This helps for cutoffs that lie outside the
					// maximum attenuation. Experimentally 0.01 is reasonable.
					if (attenuation < 0.01)
						continue;
				} else {
					// Directional light; direction is constant.
					L = l->pos_obj;
				}
				d = L.dot(*n);

				if (d >= 0) {
					// set diffuse light vertex values -- 
					// not currently faster to skip multiplies when d<0
					r = l->mat_diffuse.r * d;
					g = l->mat_diffuse.g * d;
					b = l->mat_diffuse.b * d;
				} else {
					r = 0; g = 0; b = 0; 
				}

				if (l->flags & LIGHT_SPECULAR) {
					// Normalize sum of unit light and unit viewer vector.
					// FIXME normalize need only be done once if light is directional
					if (0 && local_viewer) {  // XXX works (I think) but disabled for now for speed
						// Local viewer: get unit viewer vector
						VD.x = local_viewer_obj.x - v_obj->x;
						VD.y = local_viewer_obj.y - v_obj->y;
						VD.z = local_viewer_obj.z - v_obj->z;
						VD.normalize_self();
					}
					Vec3 s = L + VD; // specular sum
					s.normalize_self();
					d = s.dot(*n);
					if (d > 0.0f) {
						GET_SHINE_TAB_ENTRY(rctx._ShineTable[0], d, d);
						r += l->mat_specular.r * d;
						g += l->mat_specular.g * d;
						b += l->mat_specular.b * d;
					}
				}
				r += l->mat_ambient.r;
				g += l->mat_ambient.g;
				b += l->mat_ambient.b;
				if (attenuation != 1.0f)
					r *= attenuation; g*= attenuation; b*=attenuation;
				sum.r += r; sum.g += g; sum.b += b;
			}

			sum.clamp();                           // Clamp final rgb values to 1.0
			// Draw the vertex.
			// I currently transform here because we're culling vertices--
			// using mat_transform instead is equally as fast, with cubes, because
			// in general most vertices are attached to a visible cube facet anyway.
			plx_vert_indm3(&rctx.dr, (vtx==m.nfacevtx-1) ? PVR_CMD_VERTEX_EOL : PVR_CMD_VERTEX, 
			             v_obj->x, v_obj->y, v_obj->z, a0 | plx_pack_color3(sum));
		}
		// next face; next normal
		n++; f++;
	}

	m.invalidate_lighting();
}
