#include <vector>
#include <math.h>
#include "vec3.h"
#include "shapes.h"
#include "model.h"
#include "global.h"

void Sphere::sphere() { /* {{{ */
	Model& m = model_;
	float r = radius_;

	m.nvert = stacks_ * (slices_+2);
	VectorF *v  = m.data = new VectorF[m.nvert];
	VectorF *vt = m.data_xform = new VectorF[m.nvert];
	Vec3 *n     = m.normals = new Vec3[m.nvert];

	typedef Vec3::hash_map VM;
	VM vm;
	int idx = 0, nvert = 0;
	/* Iterate over stacks */
	for (int i=0; i<stacks_; i++ ) {
		float pitch = 2*M_PI * ( (float)i/(float)stacks_ );
		float pitch2 = 2*M_PI * ( (float)(i+1)/(float)stacks_ );

		/* Iterate over slices: each entire stack will be one
		   long triangle strip. */
		for (int j=0; j<=slices_/2; j++ ) {
			float yaw = 2*M_PI * ( (float)j/(float)slices_ );

			/* x, y+1 */
			v->x = r * fcos( yaw ) * fcos( pitch2 );
			v->y = r * fsin( pitch2 );
			v->z = r * fsin( yaw ) * fcos( pitch2 );
			vt->flags = PVR_CMD_VERTEX;

			// Determine vertex buffer index.
			Vec3 vtx(v->x, v->y, v->z);
			VM::iterator it = vm.find(vtx);
			if (it == vm.end())
				vm[vtx] = v->flags = idx++;  // Not in map; assign a new vertex index
			else
				v->flags = (*it).second;     // Found; use available index
			
			/* find unit normal vector -- normal in sphere is just [x,y,z] */ 
			n->x = v->x; n->y = v->y; n->z = v->z;
			n->normalize_self();
			n++; v++; vt++; nvert++;

			/* x, y */
			v->x = r * fcos( yaw ) * fcos( pitch );
			v->y = r * fsin( pitch );
			v->z = r * fsin( yaw ) * fcos( pitch );
			vt->flags = (j == slices_/2) ? PVR_CMD_VERTEX_EOL : PVR_CMD_VERTEX;
			/* Store vertex index (into lit_color) in untransformed object's vertex flags */
			
			// Determine vertex buffer index.
			vtx = Vec3(v->x, v->y, v->z);
			it = vm.find(vtx);
			if (it == vm.end())
				vm[vtx] = v->flags = idx++;  // Not in map; assign a new vertex index
			else
				v->flags = (*it).second;     // Found; use available index

			/* find unit normal vector -- normal in sphere is [x,y,z] */ 
			n->x = v->x; n->y = v->y; n->z = v->z;
			n->normalize_self();
			n++; v++; vt++; nvert++;
		}
	}
	printf("sphere: found %d unique vertices of %d total\n", idx, nvert);
	assert(nvert == m.nvert);
	m.ndups = m.nvert - idx;  // Number of duplicate vertices, for info purposes
	// XXX Watch out--removing the next 2 printfs causes a major performance hit--
	// either data or code must be aligned differently
	printf("number of indices %d\n", idx);
	//if (idx & 7)
	//	idx += 8 - (idx & 7);  // Round index length up to nearest 32 bytes
	printf("number of indices padded for sq_clr (8 dword alignment) %d\n", idx);
	m.lit_color = new uint32[idx /*+ 8*/];  /* add 32 bytes for 32-byte alignment */
	//m.lit_color = (uint32 *)(((uint32)m.lit_color & ~31) + 32); /* align to 32 bytes */
	m.nindices = idx;
	m.invalidate_lighting();
}
/* }}} */

void Rectangle::rectangle() {
	Model& m = model_;
	float h = h_/2; float w = w_/2; float d = 1.0/2;
	int idx=0;

	m.nvert = 4;
	VectorF *v  = m.data = new VectorF[m.nvert];
	VectorF *vt = m.data_xform = new VectorF[m.nvert];
	Vec3 *n     = m.normals = new Vec3[m.nvert];
	Vec3 nrm1(0,0,1);
	// Build face.  Model::draw requires the lighting buffer for now.
	v->x = -w; v->y = -h; v->z = d; *n = nrm1; v->flags = idx++; vt->flags = PVR_CMD_VERTEX; v++, vt++; n++;
	v->x = -w; v->y =  h; v->z = d; *n = nrm1; v->flags = idx++; vt->flags = PVR_CMD_VERTEX; v++, vt++; n++;
	v->x =  w; v->y = -h; v->z = d; *n = nrm1; v->flags = idx++; vt->flags = PVR_CMD_VERTEX; v++, vt++; n++;
	v->x =  w; v->y =  h; v->z = d; *n = nrm1; v->flags = idx++; vt->flags = PVR_CMD_VERTEX_EOL; v++, vt++; n++;

	m.lit_color = new uint32[m.nvert];
	m.nindices = m.nvert;
	m.invalidate_lighting();
}

void Cube::create() {
	Model& m = model_;
	float h = h_/2; float w = w_/2; float d = d_/2;
	int idx=0;

	const int FACES = 6; const int VERTS = 4;
	Vec3 V[] = { Vec3(-w,-h,d), Vec3(-w,h,d), Vec3(w,-h,d), Vec3(w,h,d),
	             Vec3(-w,-h,-d), Vec3(-w,h,-d), Vec3(w,-h,-d), Vec3(w,h,-d) };
	static uint8 faces[FACES][VERTS] = {
		{ 0, 1, 2, 3 }, // front
		{ 4, 5, 0, 1 }, // left
		{ 2, 3, 6, 7 }, // right
		{ 1, 5, 3, 7 }, // top
		{ 4, 0, 6, 2 }, // bottom
		{ 4, 5, 6, 7 }  // back
	};
	static Vec3 N[FACES] = { 
		Vec3(0,0,1), Vec3(-1,0,0), Vec3(1,0,0), Vec3(0,1,0), Vec3(0,-1,0), Vec3(0,0,-1)
	};

	m.nvert = FACES*VERTS;
	VectorF *v  = m.data = new VectorF[m.nvert];
	VectorF *vt = m.data_xform = new VectorF[m.nvert];
	Vec3 *n     = m.normals = new Vec3[m.nvert];

	for (int face = 0; face < FACES; face++) {
		for (int vtx=0; vtx < VERTS; vtx++) {
			Vec3& vc = V[faces[face][vtx]];
			v->x = vc.x + xo_; v->y = vc.y + yo_; v->z = vc.z + zo_; *n = N[face];
			v->flags = idx++; vt->flags = (vtx == VERTS-1) ? PVR_CMD_VERTEX_EOL : PVR_CMD_VERTEX;
			v++, vt++, n++;
		}
	}
	m.lit_color = new uint32[m.nvert];
	m.nindices = m.nvert;
	m.invalidate_lighting();
}

void QuadCube::create() {
	QuadModel& m = model_;
	float h = h_/2; float w = w_/2; float d = d_/2;

	const int FACES = 6; const int VERTS = QuadModel::nfacevtx;
	static Vec3 V[8] = { Vec3(-1,-1,1), Vec3(-1,1,1), Vec3(1,-1,1), Vec3(1,1,1),
	             Vec3(-1,-1,-1), Vec3(-1,1,-1), Vec3(1,-1,-1), Vec3(1,1,-1) };
	static uint8 faces[FACES][VERTS] = {
		{ 0, 1, 2, 3 }, // front
		{ 4, 5, 0, 1 }, // left
		{ 2, 3, 6, 7 }, // right
		{ 1, 5, 3, 7 }, // top
		{ 4, 0, 6, 2 }, // bottom
		{ 4, 5, 6, 7 }  // back
	};
	static Vec3 N[FACES] = { 
		Vec3(0,0,1), Vec3(-1,0,0), Vec3(1,0,0), Vec3(0,1,0), Vec3(0,-1,0), Vec3(0,0,-1)
	};
	assert(sizeof(V)/sizeof(V[0]) == 8);
	m.nvert = 8;
	m.nfaces = FACES;
	Vec3 *v  = m.data = new Vec3[m.nvert];
	Vec3 *n  = m.normals = new Vec3[m.nfaces];
	QuadModel::facedata *f  = m.faces = new QuadModel::facedata[m.nfaces];
	m.data_xform = new Vec3[m.nvert];

	// Initialize vertices
	for (int i = 0; i < m.nvert; i++) {
		Vec3& vc = V[i];
		v->x = vc.x*w + xo_; v->y = vc.y*h + yo_; v->z = vc.z*d + zo_;
		v++;
	}

	for (int i = 0; i < FACES; i++) {
	}

	// Initialize face indices and normals
	for (int face = 0; face < FACES; face++) {
		for (int vtx=0; vtx < VERTS; vtx++) {
			f->vtx[vtx] = faces[face][vtx];
		}
		// Next normal
		*n++ = N[face];
		// Initialize facet normal position programmatically
		// and precompute P*N for culling.
		Vec3 POS;
		POS.x = N[face].x * w + xo_;
		POS.y = N[face].y * h + yo_;
		POS.z = N[face].z * d + zo_;
		f->pn = POS.dot(N[face]);
		// Next face
		f++;
	}
	m.invalidate_lighting();
}


