/*      
   Tsunami for KallistiOS ##version##
        
   vector.h

   (c)2002 Dan Potter

   $Id: vector.h,v 1.2 2003/04/24 02:57:20 bardtx Exp $
*/

#ifndef __VEC3_H
#define __VEC3_H

#include <kos/vector.h>
#include <assert.h>

#include <hash_map>
#include <string>
#include "math.h"
class Vec4;

// Predeclare these, since they're used in the Vec3 class.
class Vec3;
inline Vec3 operator*(float, const Vec3&);
inline Vec3 operator/(float, const Vec3&);

class Vec3 {
public:
	Vec3() {}

	Vec3(float ix, float iy, float iz)
		: x(ix), y(iy), z(iz) { }

	/// Return one of the vector elements array-style
	float operator[](int i) const {
		if (i == 0)
			return x;
		else if (i == 1)
			return y;
		else if (i == 2)
			return z;
		else {
			assert_msg(false, "Point::operator[] Invalid index");
			return 0.0f;
		}
	}

	/// Compare two vectors for equality
	bool operator==(const Vec3 & other) const {
		return x == other.x && y == other.y && z == other.z;
	}

	/// Compare two vectors for inequality
	bool operator!=(const Vec3 & other) const {
		return !(*this == other);
	}

	/// Add two vectors
	Vec3 operator+(const Vec3 & other) const {
		return Vec3(x + other.x, y + other.y, z + other.z);
	}

	/// Subtract two vectors
	Vec3 operator-(const Vec3 & other) const {
		return Vec3(x - other.x, y - other.y, z - other.z);
	}

	/// Unary minus
	Vec3 operator-() const {
		return Vec3(-x, -y, -z);
	}

	/// Multiply by a scalar
	Vec3 operator*(float s) const {
		return Vec3(x * s, y * s, z * s);
	}
	/// Divide by a scalar
	Vec3 operator/(float s) const {
		return operator*(1.0 / s);
	}

	/// Inline add two vectors
	Vec3 & operator+=(const Vec3 & other) {
		x += other.x;
		y += other.y;
		z += other.z;
		return *this;
	}

	/// Inline subtract two vectors
	Vec3 & operator-=(const Vec3 & other) {
		x -= other.x;
		y -= other.y;
		z -= other.z;
		return *this;
	}

	// Inline multiply by a scalar
	Vec3 & operator*=(float s) {
		x *= s;
		y *= s;
		z *= s;
		return *this;
	}

	/// Get a C vector_t struct out of it
	operator vector_t() const {
		vector_t v = { x, y, z, 1.0 };
		return v;
	}

	/// Dot product with another vector.
	/// NOTE: Only takes x,y,z into account.
	float dot(const Vec3 & other) const {
		return (x * other.x)
			+ (y * other.y)
			+ (z * other.z);
	}

	/// Cross product with another vector
	/// NOTE: Only takes x,y,z into account.
	Vec3 cross(const Vec3 & other) const {
		return Vec3(
			y * other.z - z*other.y,
			z * other.x - x*other.z,
			x * other.y - y*other.x);
	}

	float length_2() const {
		return dot(*this);
	}

	/// Get the length/magnitude of the vector
	float length() const {
		return SQRT(x*x+y*y+z*z);
	}

	/// Inverse length (used for normalize)
	float inv_length() const {
		return RSQRT(x*x+y*y+z*z);
	}

	/// Return a new normalized vector
	Vec3 normalize() const {
		float l = inv_length();
		return Vec3(x*l, y*l, z*l);
	}
	
	/// Normalize self
	Vec3& normalize_self() {
		*this *= inv_length();
		return *this;
		// Could use *this *= inv_length() if we want to include w
	}

	/// Reflect this vector V across a normal N.  V should be incoming.
	/// Formula for outgoing vector A: Vr = 2(A.N)*N - A
	/// Note A = -V.  Thus Vr = 2((-V).N)*N - (-V) == 2(-(V.N))*N + V = V - 2(V.N)*N
	Vec3 reflect(const Vec3 &N) const {
		const Vec3& V = *this;
		return Vec3(V - 2*V.dot(N)*N);
	}

	std::string to_str() const {
		char buf[80];
		sprintf(buf, "[%.3f %.3f %.3f]", x, y, z);
		return std::string(buf);  // damn not having iostreams
	}

public:
	float	x, y, z;
	struct hash {
		size_t operator()(const Vec3& v) const { 
			return static_cast<size_t>(
					(int)(v.x*100) + 
					(int)(v.y*100)*1000 +
					(int)(v.z*100)*1000000
			);
		}
	};
	typedef std::hash_map<Vec3, int, hash> hash_map;

};

// Allow operations with scalar on the left.
inline Vec3 operator*(float s, const Vec3& v) { return v * s; }
inline Vec3 operator/(float s, const Vec3& v) { return v / s; }

#endif	/* __VEC3_H */

