/*      
   Tsunami for KallistiOS ##version##
        
   vector.h

   (c)2002 Dan Potter

   $Id: vector.h,v 1.2 2003/04/24 02:57:20 bardtx Exp $
*/

#ifndef __VEC4_H
#define __VEC4_H

#include <kos/vector.h>
#include <assert.h>
#include <string>
#include "vec3.h"
#include "math.h"

#ifdef _arch_dreamcast
#	include <dc/fmath.h>
#else
#	include <math.h>
#endif

class Vec4 {
public:
	Vec4() {}

	Vec4(float ix, float iy, float iz, float iw)
		: x(ix), y(iy), z(iz), w(iw) { }

	Vec4(const Vec3& v, float iw) : x(v.x), y(v.y), z(v.z), w(iw) { }

	operator Vec3() const {return Vec3(x,y,z);}
	//operator Vec3&() const {return *this;}

	/// Return one of the vector elements array-style
	float operator[](int i) const {
		if (i == 0)
			return x;
		else if (i == 1)
			return y;
		else if (i == 2)
			return z;
		else if (i == 3)
			return w;
		else {
			assert_msg(false, "Point::operator[] Invalid index");
			return 0.0f;
		}
	}

	/// Compare two vectors for equality
	bool operator==(const Vec4 & other) const {
		return x == other.x && y == other.y && z == other.z && w == other.w;
	}

	/// Compare two vectors for inequality
	bool operator!=(const Vec4 & other) const {
		return !(*this == other);
	}

	/// Add two vectors
	Vec4 operator+(const Vec4 & other) const {
		return Vec4(x + other.x, y + other.y, z + other.z, w + other.w);
	}

	/// Subtract two vectors
	Vec4 operator-(const Vec4 & other) const {
		return Vec4(x - other.x, y - other.y, z - other.z, w - other.w);
	}

	/// Unary minus
	Vec4 operator-() const {
		return Vec4(-x, -y, -z, w);
	}

	/// Multiply by a scalar
	Vec4 operator*(float s) const {
		return Vec4(x * s, y * s, z * s, w * s);
	}

	// Only assign 3 of 4 values (ignore w)
	Vec4& assign3(const Vec4& v) {
		x = v.x; y = v.y; z = v.z;
		return *this;
	}

	/// Inline add two vectors
	Vec4 & operator+=(const Vec4 & other) {
		x += other.x;
		y += other.y;
		z += other.z;
		w += other.w;
		return *this;
	}

	/// Inline subtract two vectors
	Vec4 & operator-=(const Vec4 & other) {
		x -= other.x;
		y -= other.y;
		z -= other.z;
		w -= other.w;
		return *this;
	}

	// Inline multiply by a scalar
	Vec4 & operator*=(float s) {
		x *= s;
		y *= s;
		z *= s;
		w *= s;
		return *this;
	}

	/// Get a C vector_t struct out of it
	operator vector_t() const {
		vector_t v = { x, y, z, w };
		return v;
	}

	/// Dot product with another vector.
	/// NOTE: Only takes x,y,z into account.
	float dot(const Vec4 & other) const {
		return (x * other.x)
			+ (y * other.y)
			+ (z * other.z);
	}
	float dot(const Vec3 & other) const {
		return (x * other.x)
			+ (y * other.y)
			+ (z * other.z);
	}

	/// Cross product with another vector
	/// NOTE: Only takes x,y,z into account.
	Vec4 cross(const Vec4 & other) const {
		return Vec4(
			y * other.z - z*other.y,
			z * other.x - x*other.z,
			x * other.y - y*other.x, w);
	}

	/// Get the length/magnitude of the vector
	float length() const {
#ifdef _arch_dreamcast
		return fsqrt(x*x+y*y+z*z);
#else
		return (float)sqrt(x*x+y*y+z*z);
#endif
	}

	float inv_length() const {
#ifdef _arch_dreamcast
		return frsqrt(x*x+y*y+z*z);
#else
		return 1.0f / length();
#endif

	}

	/// Return a new normalized vector
	Vec4 normalize() const {
		float l = inv_length();
		return Vec4(x*l, y*l, z*l, w);
	}
	
	/// Normalize self
	Vec4& normalize_self() {
		float l = inv_length();
		x *= l; y *= l; z *= l;
		return *this;
		// Could use *this *= inv_length() if we want to include w
	}

	Vec4& transform_nodiv() {
		mat_trans_nodiv(x, y, z, w);
		return *this;
	}

	std::string to_str() const {
		char buf[80];
		sprintf(buf, "[%.3f %.3f %.3f %.3f]", x, y, z, w);
		return std::string(buf);  // damn not having iostreams
	}

public:
	float	x, y, z, w;
};

#endif	/* __VEC4_H */

